"""
$(TYPEDEF)

Hdiv-conforming vector-valued (ncomponents = edim) lowest-order Raviart-Thomas space

allowed ElementGeometries:
- Triangle2D
- Quadrilateral2D
- Tetrahedron3D
- Hexahedron3D
"""
abstract type HDIVRT0{edim} <: AbstractHdivFiniteElement where {edim<:Int} end

get_ncomponents(FEType::Type{<:HDIVRT0}) = FEType.parameters[1]
get_ndofs_on_face(FEType::Type{<:HDIVRT0}, EG::Type{<:AbstractElementGeometry}) = 1
get_ndofs_on_cell(FEType::Type{<:HDIVRT0}, EG::Type{<:AbstractElementGeometry}) = nfaces_for_geometry(EG)

get_polynomialorder(::Type{<:HDIVRT0{2}}, ::Type{<:AbstractElementGeometry1D}) = 0;
get_polynomialorder(::Type{<:HDIVRT0{2}}, ::Type{<:AbstractElementGeometry2D}) = 1;
get_polynomialorder(::Type{<:HDIVRT0{3}}, ::Type{<:AbstractElementGeometry2D}) = 0;
get_polynomialorder(::Type{<:HDIVRT0{3}}, ::Type{<:AbstractElementGeometry3D}) = 1;


function init!(FES::FESpace{FEType}) where {FEType <: HDIVRT0}
    ncomponents = get_ncomponents(FEType)
    FES.name = "RT0 (H1, $(ncomponents)d)"

    # count number of dofs
    nfaces = num_sources(FES.xgrid[FaceNodes])
    FES.ndofs = nfaces

end

function init_dofmap!(FES::FESpace{FEType}, ::Type{CellDofs}) where {FEType <: HDIVRT0}
    xCellFaces = FES.xgrid[CellFaces]
    xCellGeometries = FES.xgrid[CellGeometries]
    dofs4item = zeros(Int32,max_num_targets_per_source(xCellFaces))
    ncells = num_sources(xCellFaces)
    xCellDofs = VariableTargetAdjacency(Int32)
    nfaces4item = 0
    for cell = 1 : ncells
        nfaces4item = num_targets(xCellFaces,cell)
        for k = 1 : nfaces4item
            dofs4item[k] = xCellFaces[k,cell]
        end
        append!(xCellDofs,dofs4item[1:nfaces4item])
    end
    # save dofmap
    FES.dofmaps[CellDofs] = xCellDofs
end

function init_dofmap!(FES::FESpace{FEType}, ::Type{FaceDofs}) where {FEType <: HDIVRT0}
    nfaces = num_sources(FES.xgrid[FaceNodes])
    xFaceDofs = VariableTargetAdjacency(Int32)
    for face = 1 : nfaces
        append!(xFaceDofs,[face])
    end
    # save dofmap
    FES.dofmaps[FaceDofs] = xFaceDofs
end

function init_dofmap!(FES::FESpace{FEType}, ::Type{BFaceDofs}) where {FEType <: HDIVRT0}
    xBFaces = FES.xgrid[BFaces]
    nbfaces = length(xBFaces)
    xBFaceDofs = VariableTargetAdjacency(Int32)
    for bface = 1: nbfaces
        append!(xBFaceDofs,[xBFaces[bface]])
    end
    # save dofmap
    FES.dofmaps[BFaceDofs] = xBFaceDofs
end


function interpolate!(Target::AbstractArray{<:Real,1}, FE::FESpace{<:HDIVRT0}, exact_function!::Function; dofs = [], bonus_quadorder::Int = 1)
    # integrate normal flux of exact_function over edges
    ncomponents = get_ncomponents(eltype(FE))
    xFaceNormals = FE.xgrid[FaceNormals]
    function normalflux_eval()
        temp = zeros(Float64,ncomponents)
        function closure(result, x, face, xref)
            exact_function!(temp,x)
            result[1] = 0.0
            for j = 1 : ncomponents
                result[1] += temp[j] * xFaceNormals[j,face]
            end 
        end   
    end   
    integrate!(Target, FE.xgrid, ON_FACES, normalflux_eval(), bonus_quadorder, 1; item_dependent_integrand = true)
end


function get_basis_normalflux_on_face(::Type{<:HDIVRT0}, ::Type{<:AbstractElementGeometry})
    function closure(refbasis, xref)
        refbasis[1,1] = 1
    end
end

function get_basis_on_cell(::Type{HDIVRT0{2}}, ::Type{<:Triangle2D})
    function closure(refbasis, xref)
        refbasis[1,:] .= [xref[1], xref[2]-1.0]
        refbasis[2,:] .= [xref[1], xref[2]]
        refbasis[3,:] .= [xref[1]-1.0, xref[2]]
    end
end

function get_basis_on_cell(::Type{HDIVRT0{2}}, ::Type{<:Quadrilateral2D})
    function closure(refbasis, xref)
        refbasis[1,:] .= [0.0, xref[2]-1.0]
        refbasis[2,:] .= [xref[1], 0.0]
        refbasis[3,:] .= [0.0, xref[2]]
        refbasis[4,:] .= [xref[1]-1.0, 0.0]
    end
end

function get_basis_on_cell(::Type{HDIVRT0{3}}, ::Type{<:Tetrahedron3D})
    function closure(refbasis, xref)
        refbasis[1,:] .= 2*[xref[1], xref[2], xref[3]-1.0]
        refbasis[2,:] .= 2*[xref[1], xref[2]-1.0, xref[3]]
        refbasis[3,:] .= 2*[xref[1], xref[2], xref[3]]
        refbasis[4,:] .= 2*[xref[1]-1.0, xref[2], xref[3]]
    end
    # note: factor 2 is chosen, such that normal-flux integrated over faces is 1 again
end

function get_basis_on_cell(::Type{HDIVRT0{3}}, ::Type{<:Hexahedron3D})
    function closure(refbasis, xref)
        refbasis[1,:] .= [0.0, 0.0, xref[3]-1.0]
        refbasis[2,:] .= [0.0, xref[2]-1.0, 0.0]
        refbasis[3,:] .= [xref[1], 0.0, 0.0]
        refbasis[4,:] .= [0.0, xref[2], 0.0]
        refbasis[5,:] .= [xref[1]-1.0, 0.0, 0.0]
        refbasis[6,:] .= [0.0, 0.0, xref[3]]
    end
end


function get_coefficients_on_cell!(FE::FESpace{<:HDIVRT0}, EG::Type{<:AbstractElementGeometry})
    xCellFaceSigns = FE.xgrid[CellFaceSigns]
    nfaces = nfaces_for_geometry(EG)
    function closure(coefficients, cell)
        # multiplication with normal vector signs
        for j = 1 : nfaces,  k = 1 : size(coefficients,1)
            coefficients[k,j] = xCellFaceSigns[j,cell];
        end
        return nothing
    end
end    
